@@ -24,11 +24,13 @@ from BayesOpt4dftu.special_kpath import kpath_dict
 from vaspvis import Band
 from vaspvis.utils import get_bandgap
 
+from matplotlib import pyplot as plt
+from matplotlib import cm, gridspec
+
 # TODO: 1. SCF calculation in DFT+U missing U tags in INCAR.
 #       2. Check whether the U value has an duplicate in u.txt.
-#       3. Add header in u.txt
-#       4. Modify the BO for multi-U condition (More than 2 U values need to be optimized).
-#       5. Fix the bug that code output incorrect U in when U values are optimized for elements without the first one.
+#       3. Modify the BO for multi-U condition (More than 2 U values need to be optimized).
+#       4. Fix the bug that code output incorrect U in when U values are optimized for elements without the first one.
 
 
 def readgap(vasprun, kpoints):
@@ -218,6 +220,7 @@ class delta_band(object):
         ispin_hse, nbands_hse, nkpts_hse = self.readInfo(self.vasprun_hse)
         ispin_dftu, nbands_dftu, nkpts_dftu = self.readInfo(self.vasprun_dftu)
 
+        
         if nbands_hse != nbands_dftu:
             raise Exception('The band number of HSE and GGA+U are not match!')
 
@@ -343,6 +346,7 @@ class delta_band(object):
 
             incar = Incar.from_file('./dftu/band/INCAR')
             u = incar['LDAUU']
+
             u.append(bg)
             u.append(delta_band)
             output = ' '.join(str(x) for x in u)
@@ -357,40 +361,46 @@ class delta_band(object):
 
 
 class bayesOpt_DFTU(object):
-    def __init__(self, path, kappa=2.5, alpha_1=1, alpha_2=1):
+    def __init__(self, path, opt_u_index=(1, 1), u_range=(0, 10), kappa=2.5, alpha_1=1, alpha_2=1):
         self.input = path + 'u.txt'
         self.gap = readgap(path + '/hse/band/vasprun.xml',
                            path + '/hse/band/KPOINTS')
         self.kappa = kappa
         self.a1 = alpha_1
         self.a2 = alpha_2
+        self.opt_u_index = np.array(opt_u_index) > 0
+        self.u_range = u_range
+        self.elements = {}
 
     def loss(self, y, y_hat, delta_band, alpha_1, alpha_2):
         return -alpha_1 * (y - y_hat) ** 2 - alpha_2 * delta_band ** 2
-
-    def bo(self, opt_u_index=(1, 1), u_range=(0, 10)):
-        data = pd.read_csv(self.input, header=None,
+    
+    def get_optimizer(self):
+        data = pd.read_csv(self.input, header=0,
                            delimiter="\s", engine='python')
         num_rows, d = data.shape
-        num_variables = sum(opt_u_index)
+        num_variables = sum(self.opt_u_index)
+        if num_variables > 2:
+            raise ValueError("BO larger than 2D are not supported yet!")
         variables_string = ascii_lowercase[:num_variables]
         pbounds = {}
         if num_variables == 1:
-            pbounds[variables_string[0]] = u_range
+            pbounds[variables_string[0]] = self.u_range
         elif num_variables == 2:
             for variable in variables_string:
-                pbounds[variable] = u_range
-        utility = UtilityFunction(kind="ucb", kappa=self.kappa, xi=0)
+                pbounds[variable] = self.u_range
+        utility_function = UtilityFunction(kind="ucb", kappa=self.kappa, xi=0)
         optimizer = BayesianOptimization(
             f=None,
             pbounds=pbounds,
             verbose=2,
             random_state=1,
         )
+        
         for i in range(num_rows):
             values = list()
-            for j in range(len(opt_u_index)):
-                if opt_u_index[j]:
+            for j in range(len(self.opt_u_index)):
+                if self.opt_u_index[j]:
                     values.append(data.iloc[i][j])
             params = {}
             for (value, variable) in zip(values, variables_string):
@@ -402,21 +412,99 @@ class bayesOpt_DFTU(object):
                 params=params,
                 target=target,
             )
-        next_point_to_probe = optimizer.suggest(utility)
+        
+        return utility_function, optimizer, target
+
+    def plot_bo(self, ratio=1):
+        utility_function, optimizer, target = self.get_optimizer()
+        plot_size = len(optimizer.res)*ratio
+        opt_eles = [ele for i, ele in enumerate(self.elements) if self.opt_u_index[i]]
+
+        if sum(self.opt_u_index) == 1:
+            x = np.linspace(self.u_range[0], self.u_range[1], 10000).reshape(-1, 1)
+            x_obs = np.array([res["params"]['a'] for res in optimizer.res]).reshape(-1,1)[:plot_size]
+            y_obs = np.array([res["target"] for res in optimizer.res])[:plot_size]
+
+            mu, sigma = posterior(optimizer, x_obs, y_obs, x)
+
+            fig = plt.figure()
+            gs = gridspec.GridSpec(2, 1) 
+            axis = plt.subplot(gs[0])
+            acq = plt.subplot(gs[1])
+            axis.plot(x_obs.flatten(), y_obs, 'D', markersize=8, label=u'Observations', color='r')
+            axis.plot(x, mu, '--', color='k', label='Prediction')
+            axis.fill(np.concatenate([x, x[::-1]]), 
+                    np.concatenate([mu - 1.9600 * sigma, (mu + 1.9600 * sigma)[::-1]]),
+                alpha=.6, fc='c', ec='None', label='95% confidence interval')
+            
+            axis.set_xlim(self.u_range)
+            axis.set_ylim((None, None))
+            axis.set_ylabel('f(x)')
+
+            acq.plot(x, utility_function, label='Acquisition Function', color='purple')
+            acq.plot(x[np.argmax(utility_function)], np.max(utility_function), '*', markersize=15, 
+                    label=u'Next Best Guess', markerfacecolor='gold', markeredgecolor='k', markeredgewidth=1)
+            acq.set_xlim(self.u_range)
+            acq.set_ylim((np.min(utility_function)-0.5,np.max(utility_function)+0.5))
+            acq.set_ylabel('Acquisition')
+            acq.set_xlabel('U (eV)')
+            axis.legend(loc=4, borderaxespad=0.)
+            acq.legend(loc=4, borderaxespad=0.)
+
+            plt.savefig('1D_kappa_%s_a1_%s_a2_%s.png' %(self.kappa, self.a1, self.a2), dpi = 400)
+
+        if sum(self.opt_u_index) == 2:
+            x = y = np.linspace(self.u_range[0], self.u_range[1], 300)
+            X, Y = np.meshgrid(x, y)
+            x = X.ravel()
+            y = Y.ravel()
+            X = np.vstack([x, y]).T[:, [1, 0]]
+
+            x1_obs = np.array([[res["params"]["a"]] for res in optimizer.res])[:plot_size]
+            x2_obs = np.array([[res["params"]["b"]] for res in optimizer.res])[:plot_size]
+            y_obs = np.array([res["target"] for res in optimizer.res])[:plot_size]
+            obs = np.column_stack((x1_obs, x2_obs))
+
+            optimizer._gp.fit(obs, y_obs)
+            mu, sigma = optimizer._gp.predict(X, eval)
+
+            fig, axis = plt.subplots(1, 2, figsize=(15,5))
+            plt.subplots_adjust(wspace = 0.2)
+            
+            axis[0].plot(x1_obs, x2_obs, 'D', markersize=4, color='k', label='Observations')
+            axis[0].set_title('Gaussian Process Predicted Mean',pad=10)
+            im1 = axis[0].hexbin(y, x, C=mu, cmap=cm.jet, bins=None)
+            axis[0].axis([x.min(), x.max(), y.min(), y.max()])
+            axis[0].set_xlabel('U_%s (eV)' %opt_eles[0],labelpad=5)
+            axis[0].set_ylabel('U_%s (eV)' %opt_eles[1],labelpad=10,va='center')
+            cbar1 = plt.colorbar(im1, ax = axis[0])
+
+            utility = utility_function.utility(X, optimizer._gp, optimizer.max)
+            axis[1].plot(x1_obs, x2_obs, 'D', markersize=4, color='k', label='Observations')
+            axis[1].set_title('Acquisition Function',pad=10)
+            axis[1].set_xlabel('U_%s (eV)' %opt_eles[0],labelpad=5)
+            axis[1].set_ylabel('U_%s (eV)' %opt_eles[1],labelpad=10,va='center')
+            im3 = axis[1].hexbin(y, x, C=utility, cmap=cm.jet, bins=None)
+            axis[1].axis([x.min(), x.max(), y.min(), y.max()])
+            cbar3 = plt.colorbar(im3, ax = axis[1])
+
+            plt.savefig('2D_kappa_%s_a1_%s_a2_%s.png' %(self.kappa, self.a1, self.a2), dpi = 400)
+
+
+    def bo(self):
+        utility_function, optimizer, target = self.get_optimizer()
+        next_point_to_probe = optimizer.suggest(utility_function)
 
         points = list(next_point_to_probe.values())
-        # if num_variables == 1:
-        #     if opt_u_index[0] == 1 and opt_u_index[1] == 0:
-        #         points.append(0)
-        #     elif opt_u_index[0] == 0 and opt_u_index[1] == 1:
-        #         points.insert(0,0)
         points = [round(elem, 6) for elem in points]
+
         U = [str(x) for x in points]
         with open('input.json', 'r') as f:
             data = json.load(f)
             elements = list(data["pbe"]["ldau_luj"].keys())
-            for i in range(len(opt_u_index)):
-                if opt_u_index[i]:
+            self.elements = elements
+            for i in range(len(self.opt_u_index)):
+                if self.opt_u_index[i]:
                     try:
                         data["pbe"]["ldau_luj"][elements[i]
                                                 ]["U"] = round(float(U[i]), 6)
